@@ -1,14 +1,13 @@
 use crate::{
     fs::{Mode, OFlags, ResolveFlags},
-    path,
+    io, path,
 };
 use io_lifetimes::{AsFd, BorrowedFd, OwnedFd};
-use std::{ffi::CStr, io};
+use std::ffi::CStr;
 #[cfg(libc)]
 use {
     crate::negone_err,
-    std::convert::TryInto,
-    unsafe_io::os::posish::{AsRawFd, FromRawFd},
+    unsafe_io::os::posish::{AsRawFd, FromRawFd, RawFd},
 };
 
 #[cfg(all(libc, target_pointer_width = "32"))]
@@ -37,7 +36,7 @@ pub fn openat2<'f, Fd: AsFd<'f>, P: path::Arg>(
     resolve: ResolveFlags,
 ) -> io::Result<OwnedFd> {
     let dirfd = dirfd.as_fd();
-    let path = path.as_c_str()?;
+    let path = path.into_c_str()?;
     _openat2(dirfd, &path, oflags, mode, resolve)
 }
 
@@ -66,7 +65,7 @@ fn _openat2(
         ))?;
 
         #[allow(clippy::useless_conversion)]
-        Ok(OwnedFd::from_raw_fd(fd.try_into().unwrap()))
+        Ok(OwnedFd::from_raw_fd(fd as RawFd))
     }
 }
 
