@@ -12,6 +12,12 @@ use std::{
 };
 use unsafe_io::os::posish::{AsRawFd, FromRawFd, IntoRawFd, RawFd};
 
+/// Convert `SYS_*` constants for socketcall.
+#[cfg(target_arch = "x86")]
+pub(super) fn x86_sys(sys: u32) -> usize {
+    sys as usize
+}
+
 #[cfg(all(target_endian = "little", target_pointer_width = "32"))]
 #[inline]
 pub(super) fn lo(x: u64) -> usize {
@@ -36,6 +42,18 @@ pub(super) fn hi(x: u64) -> usize {
     (x >> 32) as usize
 }
 
+#[cfg(target_pointer_width = "32")]
+#[inline]
+pub(super) fn i64_hi(x: i64) -> usize {
+    hi(x as u64)
+}
+
+#[cfg(target_pointer_width = "32")]
+#[inline]
+pub(super) fn i64_lo(x: i64) -> usize {
+    lo(x as u64)
+}
+
 #[inline]
 pub(super) fn void_star(c: *mut c_void) -> usize {
     c as usize
