@@ -12,29 +12,32 @@ use arch::{
     syscall6, syscall6_readonly,
 };
 use conv::{
-    borrowed_fd, by_mut, by_ref, c_int, c_str, c_uint, clockid_t, loff_t, loff_t_from_u64,
-    opt_c_str, opt_mut, opt_ref, out, owned_fd, ret, ret_c_int, ret_c_uint, ret_owned_fd,
-    ret_usize, ret_void_star, slice_addr, slice_as_mut_ptr, socklen_t, umode_t, void_star,
+    borrowed_fd, by_mut, by_ref, c_int, c_str, c_uint, clockid_t, opt_c_str, opt_mut, opt_ref, out,
+    owned_fd, ret, ret_c_int, ret_c_uint, ret_owned_fd, ret_usize, ret_void_star, slice_addr,
+    slice_as_mut_ptr, socklen_t, umode_t, void_star,
 };
 use io_lifetimes::{BorrowedFd, OwnedFd};
+#[cfg(not(target_arch = "x86"))]
+use linux_raw_sys::general::{
+    __NR_accept4, __NR_bind, __NR_connect, __NR_getpeername, __NR_getsockname, __NR_getsockopt,
+    __NR_listen, __NR_recvfrom, __NR_sendto, __NR_setsockopt, __NR_shutdown, __NR_socket,
+    __NR_socketpair,
+};
 use linux_raw_sys::{
     general::{
-        __NR_accept4, __NR_bind, __NR_chdir, __NR_clock_getres, __NR_clock_gettime,
-        __NR_clock_nanosleep, __NR_close, __NR_connect, __NR_dup, __NR_exit_group, __NR_faccessat,
-        __NR_fadvise64, __NR_fallocate, __NR_fchmod, __NR_fchmodat, __NR_fcntl, __NR_fdatasync,
-        __NR_fstat, __NR_fstatfs, __NR_fsync, __NR_ftruncate, __NR_getcwd, __NR_getdents64,
-        __NR_getegid, __NR_geteuid, __NR_getgid, __NR_getpeername, __NR_getpid, __NR_getppid,
-        __NR_getsockname, __NR_getsockopt, __NR_getuid, __NR_ioctl, __NR_linkat, __NR_listen,
-        __NR_lseek, __NR_mkdirat, __NR_mmap, __NR_nanosleep, __NR_newfstatat, __NR_openat,
-        __NR_pipe, __NR_pipe2, __NR_poll, __NR_pread64, __NR_preadv, __NR_pwrite64, __NR_pwritev,
-        __NR_read, __NR_readlinkat, __NR_readv, __NR_recvfrom, __NR_renameat, __NR_sched_yield,
-        __NR_sendto, __NR_setsockopt, __NR_shutdown, __NR_socket, __NR_socketpair, __NR_symlinkat,
-        __NR_unlinkat, __NR_utimensat, __NR_write, __NR_writev,
+        __NR_chdir, __NR_clock_getres, __NR_clock_gettime, __NR_clock_nanosleep, __NR_close,
+        __NR_dup, __NR_exit_group, __NR_faccessat, __NR_fadvise64, __NR_fallocate, __NR_fchmod,
+        __NR_fchmodat, __NR_fdatasync, __NR_fsync, __NR_getcwd, __NR_getdents64, __NR_getegid,
+        __NR_geteuid, __NR_getgid, __NR_getpid, __NR_getppid, __NR_getuid, __NR_ioctl, __NR_linkat,
+        __NR_mkdirat, __NR_nanosleep, __NR_openat, __NR_pipe, __NR_pipe2, __NR_poll, __NR_pread64,
+        __NR_preadv, __NR_pwrite64, __NR_pwritev, __NR_read, __NR_readlinkat, __NR_readv,
+        __NR_renameat, __NR_sched_yield, __NR_symlinkat, __NR_unlinkat, __NR_utimensat, __NR_write,
+        __NR_writev,
     },
     general::{
         __kernel_clockid_t, __kernel_gid_t, __kernel_loff_t, __kernel_pid_t,
         __kernel_sockaddr_storage, __kernel_uid_t, sockaddr_in, sockaddr_in6, sockaddr_un,
-        socklen_t, stat, statfs64, termios, timespec, umode_t, winsize,
+        socklen_t, statfs64, termios, umode_t, winsize,
     },
     general::{
         AT_FDCWD, AT_REMOVEDIR, AT_SYMLINK_NOFOLLOW, FIONREAD, F_DUPFD, F_DUPFD_CLOEXEC, F_GETFD,
@@ -43,11 +46,11 @@ use linux_raw_sys::{
     },
     v5_11::{general::__NR_openat2, general::open_how},
     v5_4::{
-        general::statx,
         general::{
             __NR_copy_file_range, __NR_getrandom, __NR_memfd_create, __NR_preadv2, __NR_pwritev2,
             __NR_statx, __NR_userfaultfd,
         },
+        general::{__kernel_timespec, statx},
         general::{F_GETPIPE_SZ, F_GET_SEALS, F_SETPIPE_SZ},
     },
 };
@@ -57,10 +60,44 @@ use std::{
     mem::{size_of, MaybeUninit},
     os::raw::{c_char, c_int, c_uint, c_void},
 };
+#[cfg(target_arch = "x86")]
+use {
+    crate::linux_raw::conv::x86_sys,
+    linux_raw_sys::general::{
+        __NR_mmap2, __NR_socketcall, SYS_ACCEPT4, SYS_BIND, SYS_CONNECT, SYS_GETPEERNAME,
+        SYS_GETSOCKNAME, SYS_GETSOCKOPT, SYS_LISTEN, SYS_RECVFROM, SYS_SENDTO, SYS_SETSOCKOPT,
+        SYS_SHUTDOWN, SYS_SOCKET, SYS_SOCKETPAIR,
+    },
+};
 #[cfg(target_pointer_width = "32")]
 use {
-    crate::linux_raw::conv::{hi, lo},
-    linux_raw_sys::{errno::EINVAL, unistd::__NR_clock_nanosleep_time64},
+    crate::linux_raw::conv::{hi, i64_hi, i64_lo, lo},
+    linux_raw_sys::{
+        errno::EINVAL,
+        general::stat64 as stat,
+        general::timespec as __kernel_old_timespec,
+        general::CLOCK_REALTIME,
+        general::{
+            __NR__llseek, __NR_fcntl64, __NR_fstat64, __NR_fstatat64, __NR_fstatfs64,
+            __NR_ftruncate64,
+        },
+        v5_4::general::{
+            __NR_clock_getres_time64, __NR_clock_gettime64, __NR_clock_nanosleep_time64,
+            __NR_utimensat_time64,
+        },
+    },
+    std::convert::TryInto,
+};
+#[cfg(target_pointer_width = "64")]
+use {
+    conv::{loff_t, loff_t_from_u64},
+    linux_raw_sys::{
+        general::stat,
+        general::{
+            __NR_fcntl, __NR_fstat, __NR_fstatfs, __NR_ftruncate, __NR_lseek, __NR_mmap,
+            __NR_newfstatat,
+        },
+    },
 };
 
 #[inline]
@@ -130,10 +167,44 @@ pub(crate) fn openat2(
 }
 
 #[inline]
-pub(crate) fn clock_gettime(which_clock: __kernel_clockid_t) -> timespec {
+pub(crate) fn clock_gettime(which_clock: __kernel_clockid_t) -> __kernel_timespec {
+    // TODO: Linux makes clock_gettime available through the vDSO, which is
+    // faster, but more complex to locate and parse before we can call it.
+    #[cfg(target_pointer_width = "32")]
+    unsafe {
+        let mut result = MaybeUninit::uninit();
+        let _ = ret(syscall2(
+            __NR_clock_gettime64,
+            clockid_t(which_clock),
+            out(&mut result),
+        ))
+        .or_else(|err| {
+            // Ordinarily posish doesn't like to emulate system calls, but in
+            // the case of time APIs, it's specific to Linux, specific to
+            // 32-bit architectures *and* specific to old kernel versions, and
+            // it's not that hard to fix up here, so that no other code needs
+            // to worry about this.
+            if err == io::Error::NOSYS {
+                let mut old_result = MaybeUninit::<__kernel_old_timespec>::uninit();
+                let res = ret(syscall2(
+                    __NR_clock_gettime,
+                    clockid_t(which_clock),
+                    out(&mut old_result),
+                ));
+                let old_result = *old_result.as_ptr();
+                *result.as_mut_ptr() = __kernel_timespec {
+                    tv_sec: old_result.tv_sec.into(),
+                    tv_nsec: old_result.tv_nsec.into(),
+                };
+                res
+            } else {
+                Err(err)
+            }
+        });
+        result.assume_init()
+    }
+    #[cfg(target_pointer_width = "64")]
     unsafe {
-        // TODO: Linux makes clock_gettime available through the vDSO, which is
-        // faster, but more complex to locate and parse before we can call it.
         let mut result = MaybeUninit::uninit();
         syscall2(__NR_clock_gettime, clockid_t(which_clock), out(&mut result));
         result.assume_init()
@@ -141,7 +212,37 @@ pub(crate) fn clock_gettime(which_clock: __kernel_clockid_t) -> timespec {
 }
 
 #[inline]
-pub(crate) fn clock_getres(which_clock: __kernel_clockid_t) -> timespec {
+pub(crate) fn clock_getres(which_clock: __kernel_clockid_t) -> __kernel_timespec {
+    #[cfg(target_pointer_width = "32")]
+    unsafe {
+        let mut result = MaybeUninit::uninit();
+        let _ = ret(syscall2(
+            __NR_clock_getres_time64,
+            clockid_t(which_clock),
+            out(&mut result),
+        ))
+        .or_else(|err| {
+            // See the comments in `clock_gettime` about emulation.
+            if err == io::Error::NOSYS {
+                let mut old_result = MaybeUninit::<__kernel_old_timespec>::uninit();
+                let res = ret(syscall2(
+                    __NR_clock_getres,
+                    clockid_t(which_clock),
+                    out(&mut old_result),
+                ));
+                let old_result = *old_result.as_ptr();
+                *result.as_mut_ptr() = __kernel_timespec {
+                    tv_sec: old_result.tv_sec.into(),
+                    tv_nsec: old_result.tv_nsec.into(),
+                };
+                res
+            } else {
+                Err(err)
+            }
+        });
+        result.assume_init()
+    }
+    #[cfg(target_pointer_width = "64")]
     unsafe {
         let mut result = MaybeUninit::uninit();
         syscall2(__NR_clock_getres, clockid_t(which_clock), out(&mut result));
@@ -165,7 +266,7 @@ pub(crate) fn read(fd: BorrowedFd<'_>, buffer: &mut [u8]) -> io::Result<usize> {
 pub(crate) fn pread(fd: BorrowedFd<'_>, buffer: &[u8], pos: u64) -> io::Result<usize> {
     #[cfg(target_pointer_width = "32")]
     unsafe {
-        ret_usize(syscall6(
+        ret_usize(syscall5(
             __NR_pread64,
             borrowed_fd(fd),
             slice_addr(buffer),
@@ -187,16 +288,6 @@ pub(crate) fn pread(fd: BorrowedFd<'_>, buffer: &[u8], pos: u64) -> io::Result<u
 }
 
 pub(crate) fn readv(fd: BorrowedFd<'_>, bufs: &[IoSliceMut]) -> io::Result<usize> {
-    #[cfg(target_pointer_width = "32")]
-    unsafe {
-        ret_usize(syscall3(
-            __NR_readv,
-            borrowed_fd(fd),
-            slice_addr(bufs),
-            bufs.len(),
-        ))
-    }
-    #[cfg(target_pointer_width = "64")]
     unsafe {
         ret_usize(syscall3(
             __NR_readv,
@@ -210,7 +301,7 @@ pub(crate) fn readv(fd: BorrowedFd<'_>, bufs: &[IoSliceMut]) -> io::Result<usize
 pub(crate) fn preadv(fd: BorrowedFd<'_>, bufs: &[IoSliceMut], pos: u64) -> io::Result<usize> {
     #[cfg(target_pointer_width = "32")]
     unsafe {
-        ret_usize(syscall6(
+        ret_usize(syscall5(
             __NR_preadv,
             borrowed_fd(fd),
             slice_addr(bufs),
@@ -278,7 +369,7 @@ pub(crate) fn write(fd: BorrowedFd<'_>, buffer: &[u8]) -> io::Result<usize> {
 pub(crate) fn pwrite(fd: BorrowedFd<'_>, buffer: &[u8], pos: u64) -> io::Result<usize> {
     #[cfg(target_pointer_width = "32")]
     unsafe {
-        ret_usize(syscall6_readonly(
+        ret_usize(syscall5_readonly(
             __NR_pwrite64,
             borrowed_fd(fd),
             slice_addr(buffer),
@@ -301,16 +392,6 @@ pub(crate) fn pwrite(fd: BorrowedFd<'_>, buffer: &[u8], pos: u64) -> io::Result<
 
 #[inline]
 pub(crate) fn writev(fd: BorrowedFd<'_>, bufs: &[IoSlice]) -> io::Result<usize> {
-    #[cfg(target_pointer_width = "32")]
-    unsafe {
-        ret_usize(syscall6_readonly(
-            __NR_writev,
-            borrowed_fd(fd),
-            slice_addr(bufs),
-            bufs.len(),
-        ))
-    }
-    #[cfg(target_pointer_width = "64")]
     unsafe {
         ret_usize(syscall3_readonly(
             __NR_writev,
@@ -325,7 +406,7 @@ pub(crate) fn writev(fd: BorrowedFd<'_>, bufs: &[IoSlice]) -> io::Result<usize>
 pub(crate) fn pwritev(fd: BorrowedFd<'_>, bufs: &[IoSlice], pos: u64) -> io::Result<usize> {
     #[cfg(target_pointer_width = "32")]
     unsafe {
-        ret_usize(syscall6_readonly(
+        ret_usize(syscall5_readonly(
             __NR_pwritev,
             borrowed_fd(fd),
             slice_addr(bufs),
@@ -416,17 +497,17 @@ pub(crate) fn fchmod(fd: BorrowedFd<'_>, mode: umode_t) -> io::Result<()> {
 #[inline]
 pub(crate) fn seek(fd: BorrowedFd<'_>, offset: i64, whence: c_uint) -> io::Result<u64> {
     #[cfg(target_pointer_width = "32")]
-    {
+    unsafe {
         let mut result = MaybeUninit::uninit();
         ret(syscall5(
             __NR__llseek,
             borrowed_fd(fd),
-            hi(offset),
-            lo(offset),
+            i64_hi(offset),
+            i64_lo(offset),
             out(&mut result),
             c_uint(whence),
         ))
-        .map(|()| unsafe { result.assume_init() })
+        .map(|()| result.assume_init())
     }
     #[cfg(target_pointer_width = "64")]
     unsafe {
@@ -462,6 +543,19 @@ pub(crate) fn ftruncate(fd: BorrowedFd<'_>, length: u64) -> io::Result<()> {
 
 #[inline]
 pub(crate) fn fallocate(fd: BorrowedFd, mode: c_int, offset: u64, len: u64) -> io::Result<()> {
+    #[cfg(target_pointer_width = "32")]
+    unsafe {
+        ret(syscall6_readonly(
+            __NR_fallocate,
+            borrowed_fd(fd),
+            c_int(mode),
+            hi(offset),
+            lo(offset),
+            hi(len),
+            lo(len),
+        ))
+    }
+    #[cfg(target_pointer_width = "64")]
     unsafe {
         ret(syscall4_readonly(
             __NR_fallocate,
@@ -475,6 +569,36 @@ pub(crate) fn fallocate(fd: BorrowedFd, mode: c_int, offset: u64, len: u64) -> i
 
 #[inline]
 pub(crate) fn fadvise(fd: BorrowedFd<'_>, pos: u64, len: u64, advice: c_int) -> io::Result<()> {
+    // On arm and powerpc, the system calls are reordered so that the len and
+    // pos argument pairs are aligned.
+    #[cfg(any(target_arch = "arm", target_arch = "powerpc"))]
+    unsafe {
+        ret(syscall6_readonly(
+            __NR_fadvise64,
+            borrowed_fd(fd),
+            c_int(advice),
+            hi(pos),
+            lo(pos),
+            hi(len),
+            lo(len),
+        ))
+    }
+    #[cfg(all(
+        target_pointer_width = "32",
+        not(any(target_arch = "arm", target_arch = "powerpc"))
+    ))]
+    unsafe {
+        ret(syscall6_readonly(
+            __NR_fadvise64,
+            borrowed_fd(fd),
+            hi(pos),
+            lo(pos),
+            hi(len),
+            lo(len),
+            c_int(advice),
+        ))
+    }
+    #[cfg(target_pointer_width = "64")]
     unsafe {
         ret(syscall4_readonly(
             __NR_fadvise64,
@@ -572,7 +696,7 @@ pub(crate) fn lstat(filename: &CStr) -> io::Result<stat> {
     #[cfg(target_pointer_width = "32")]
     unsafe {
         let mut result = MaybeUninit::uninit();
-        ret(syscall3(
+        ret(syscall4(
             __NR_fstatat64,
             c_int(AT_FDCWD),
             c_str(filename),
@@ -621,10 +745,10 @@ pub(crate) fn fstatfs(fd: BorrowedFd<'_>) -> io::Result<statfs64> {
     #[cfg(target_pointer_width = "32")]
     unsafe {
         let mut result = MaybeUninit::uninit();
-        ret(syscall2(
+        ret(syscall3(
             __NR_fstatfs64,
             borrowed_fd(fd),
-            size_of::<statfs>(),
+            size_of::<statfs64>(),
             out(&mut result),
         ))
         .map(|()| result.assume_init())
@@ -1131,7 +1255,7 @@ pub(crate) fn socket(family: c_uint, typ: c_uint, protocol: c_uint) -> io::Resul
     unsafe {
         ret_owned_fd(syscall2_readonly(
             __NR_socketcall,
-            SYS_SOCKET,
+            x86_sys(SYS_SOCKET),
             slice_addr(&[c_uint(family), c_uint(typ), c_uint(protocol), 0, 0, 0]),
         ))
     }
@@ -1163,7 +1287,7 @@ pub(crate) fn socketpair(
         let mut result = MaybeUninit::<[OwnedFd; 2]>::uninit();
         ret(syscall2(
             __NR_socketcall,
-            SYS_SOCKETPAIR,
+            x86_sys(SYS_SOCKETPAIR),
             slice_addr(&[
                 c_uint(family),
                 c_uint(typ),
@@ -1204,16 +1328,16 @@ pub(crate) fn accept(
         let mut storage = MaybeUninit::uninit();
         let fd = ret_owned_fd(syscall2(
             __NR_socketcall,
-            SYS_ACCEPT4,
+            x86_sys(SYS_ACCEPT4),
             slice_addr(&[
-                borrwed_fd(fd),
+                borrowed_fd(fd),
                 out(&mut storage),
                 by_mut(&mut addrlen),
                 c_uint(flags),
                 0,
                 0,
             ]),
-        ));
+        ))?;
         Ok((fd, storage.assume_init(), addrlen))
     }
 }
@@ -1228,8 +1352,8 @@ pub(crate) fn shutdown(fd: BorrowedFd<'_>, how: c_uint) -> io::Result<()> {
     unsafe {
         ret(syscall2_readonly(
             __NR_socketcall,
-            SYS_SHUTDOWN,
-            &[borrwed_fd(fd), c_uint(how), 0, 0, 0, 0],
+            x86_sys(SYS_SHUTDOWN),
+            slice_addr(&[borrowed_fd(fd), c_uint(how), 0, 0, 0, 0]),
         ))
     }
 }
@@ -1259,9 +1383,9 @@ pub(crate) fn setsockopt(
     unsafe {
         ret_c_int(syscall2_readonly(
             __NR_socketcall,
-            SYS_SETSOCKOPT,
+            x86_sys(SYS_SETSOCKOPT),
             slice_addr(&[
-                borrwed_fd(fd),
+                borrowed_fd(fd),
                 c_int(level),
                 c_int(name),
                 slice_addr(value),
@@ -1297,15 +1421,15 @@ pub(crate) fn getsockopt(
     unsafe {
         ret_c_int(syscall2(
             __NR_socketcall,
-            SYS_GETSOCKOPT,
-            &[
-                borrwed_fd(fd),
+            x86_sys(SYS_GETSOCKOPT),
+            slice_addr(&[
+                borrowed_fd(fd),
                 c_int(level),
                 c_int(name),
                 slice_addr(value),
                 by_mut(optlen),
                 0,
-            ],
+            ]),
         ))
     }
 }
@@ -1339,9 +1463,9 @@ pub(crate) fn sendto(
     unsafe {
         ret_usize(syscall2_readonly(
             __NR_socketcall,
-            SYS_SENDTO,
+            x86_sys(SYS_SENDTO),
             slice_addr(&[
-                borrwed_fd(fd),
+                borrowed_fd(fd),
                 slice_addr(buf),
                 buf.len(),
                 c_uint(flags),
@@ -1381,9 +1505,9 @@ pub(crate) fn recvfrom(
     unsafe {
         ret_usize(syscall2(
             __NR_socketcall,
-            SYS_RECVFROM,
+            x86_sys(SYS_RECVFROM),
             slice_addr(&[
-                borrwed_fd(fd),
+                borrowed_fd(fd),
                 slice_as_mut_ptr(buf),
                 buf.len(),
                 c_uint(flags),
@@ -1413,8 +1537,8 @@ pub(crate) fn getpeername(
     unsafe {
         ret(syscall2(
             __NR_socketcall,
-            SYS_GETPEERNAME,
-            slice_addr(&[borrwed_fd(fd), by_mut(addr), by_mut(addrlen), 0, 0, 0]),
+            x86_sys(SYS_GETPEERNAME),
+            slice_addr(&[borrowed_fd(fd), by_mut(addr), by_mut(addrlen), 0, 0, 0]),
         ))
     }
 }
@@ -1438,7 +1562,7 @@ pub(crate) fn getsockname(
     unsafe {
         ret(syscall2(
             __NR_socketcall,
-            SYS_GETSOCKNAME,
+            x86_sys(SYS_GETSOCKNAME),
             slice_addr(&[borrowed_fd(fd), by_mut(addr), by_mut(addrlen), 0, 0, 0]),
         ))
     }
@@ -1459,9 +1583,9 @@ pub(crate) fn bind_un(fd: BorrowedFd<'_>, addr: &sockaddr_un) -> io::Result<()>
     unsafe {
         ret(syscall2_readonly(
             __NR_socketcall,
-            SYS_BIND,
+            x86_sys(SYS_BIND),
             slice_addr(&[
-                borrwed_fd(fd),
+                borrowed_fd(fd),
                 by_ref(addr),
                 size_of::<sockaddr_un>(),
                 0,
@@ -1487,9 +1611,9 @@ pub(crate) fn bind_in(fd: BorrowedFd<'_>, addr: &sockaddr_in) -> io::Result<()>
     unsafe {
         ret(syscall2_readonly(
             __NR_socketcall,
-            SYS_BIND,
+            x86_sys(SYS_BIND),
             slice_addr(&[
-                borrwed_fd(fd),
+                borrowed_fd(fd),
                 by_ref(addr),
                 size_of::<sockaddr_in>(),
                 0,
@@ -1515,9 +1639,9 @@ pub(crate) fn bind_in6(fd: BorrowedFd<'_>, addr: &sockaddr_in6) -> io::Result<()
     unsafe {
         ret(syscall2_readonly(
             __NR_socketcall,
-            SYS_BIND,
+            x86_sys(SYS_BIND),
             slice_addr(&[
-                borrwed_fd(fd),
+                borrowed_fd(fd),
                 by_ref(addr),
                 size_of::<sockaddr_in6>(),
                 0,
@@ -1543,9 +1667,9 @@ pub(crate) fn connect_un(fd: BorrowedFd<'_>, addr: &sockaddr_un) -> io::Result<(
     unsafe {
         ret(syscall2_readonly(
             __NR_socketcall,
-            SYS_CONNECT,
+            x86_sys(SYS_CONNECT),
             slice_addr(&[
-                borrwed_fd(fd),
+                borrowed_fd(fd),
                 by_ref(addr),
                 size_of::<sockaddr_un>(),
                 0,
@@ -1571,9 +1695,9 @@ pub(crate) fn connect_in(fd: BorrowedFd<'_>, addr: &sockaddr_in) -> io::Result<(
     unsafe {
         ret(syscall2_readonly(
             __NR_socketcall,
-            SYS_CONNECT,
+            x86_sys(SYS_CONNECT),
             slice_addr(&[
-                borrwed_fd(fd),
+                borrowed_fd(fd),
                 by_ref(addr),
                 size_of::<sockaddr_in>(),
                 0,
@@ -1599,9 +1723,9 @@ pub(crate) fn connect_in6(fd: BorrowedFd<'_>, addr: &sockaddr_in6) -> io::Result
     unsafe {
         ret(syscall2_readonly(
             __NR_socketcall,
-            SYS_CONNECT,
+            x86_sys(SYS_CONNECT),
             slice_addr(&[
-                borrwed_fd(fd),
+                borrowed_fd(fd),
                 by_ref(addr),
                 size_of::<sockaddr_in6>(),
                 0,
@@ -1626,8 +1750,8 @@ pub(crate) fn listen(fd: BorrowedFd<'_>, backlog: c_int) -> io::Result<()> {
     unsafe {
         ret(syscall2_readonly(
             __NR_socketcall,
-            SYS_LISTEN,
-            slice_addr(&[borrwed_fd(fd), c_int(backlog), 0, 0, 0, 0]),
+            x86_sys(SYS_LISTEN),
+            slice_addr(&[borrowed_fd(fd), c_int(backlog), 0, 0, 0, 0]),
         ))
     }
 }
@@ -1663,7 +1787,7 @@ pub(crate) unsafe fn mmap(
             borrowed_fd(fd),
             (offset / 4096)
                 .try_into()
-                .map_err(|_| io::Error::from_raw_os_error(EINVAL))?,
+                .map_err(|_| io::Error(EINVAL as _))?,
         ))
     }
     #[cfg(target_pointer_width = "64")]
@@ -1684,7 +1808,7 @@ pub(crate) unsafe fn mmap(
 pub(crate) fn utimensat(
     dirfd: BorrowedFd<'_>,
     pathname: Option<&CStr>,
-    utimes: &[timespec; 2],
+    utimes: &[__kernel_timespec; 2],
     flags: c_uint,
 ) -> io::Result<()> {
     #[cfg(target_pointer_width = "32")]
@@ -1696,6 +1820,36 @@ pub(crate) fn utimensat(
             slice_addr(utimes),
             c_uint(flags),
         ))
+        .or_else(|err| {
+            // See the comments in `clock_gettime` about emulation.
+            if err == io::Error::NOSYS {
+                let old_utimes = [
+                    __kernel_old_timespec {
+                        tv_sec: utimes[0]
+                            .tv_sec
+                            .try_into()
+                            .map_err(|_| io::Error(EINVAL as _))?,
+                        tv_nsec: utimes[0].tv_nsec as _,
+                    },
+                    __kernel_old_timespec {
+                        tv_sec: utimes[1]
+                            .tv_sec
+                            .try_into()
+                            .map_err(|_| io::Error(EINVAL as _))?,
+                        tv_nsec: utimes[1].tv_nsec as _,
+                    },
+                ];
+                ret(syscall4_readonly(
+                    __NR_utimensat,
+                    borrowed_fd(dirfd),
+                    opt_c_str(pathname),
+                    slice_addr(&old_utimes),
+                    c_uint(flags),
+                ))
+            } else {
+                Err(err)
+            }
+        })
     }
     #[cfg(target_pointer_width = "64")]
     unsafe {
@@ -1710,46 +1864,123 @@ pub(crate) fn utimensat(
 }
 
 #[inline]
-pub(crate) fn nanosleep(req: &timespec) -> Result<(), timespec> {
+pub(crate) fn nanosleep(req: &__kernel_timespec) -> io::Result<Option<__kernel_timespec>> {
+    #[cfg(target_pointer_width = "32")]
+    unsafe {
+        let mut rem = MaybeUninit::<__kernel_timespec>::uninit();
+        match ret(syscall4(
+            __NR_clock_nanosleep_time64,
+            clockid_t(CLOCK_REALTIME as _),
+            c_int(0),
+            by_ref(req),
+            out(&mut rem),
+        ))
+        .or_else(|err| {
+            // See the comments in `clock_gettime` about emulation.
+            if err == io::Error::NOSYS {
+                let old_req = __kernel_old_timespec {
+                    tv_sec: req.tv_sec.try_into().map_err(|_| io::Error(EINVAL as _))?,
+                    tv_nsec: req.tv_nsec as _,
+                };
+                let mut old_rem = MaybeUninit::<__kernel_old_timespec>::uninit();
+                let res = ret(syscall2(
+                    __NR_nanosleep,
+                    by_ref(&old_req),
+                    out(&mut old_rem),
+                ));
+                let old_rem = *old_rem.as_ptr();
+                *rem.as_mut_ptr() = __kernel_timespec {
+                    tv_sec: old_rem.tv_sec.into(),
+                    tv_nsec: old_rem.tv_nsec.into(),
+                };
+                res
+            } else {
+                Err(err)
+            }
+        }) {
+            Ok(()) => Ok(None),
+            Err(io::Error::INTR) => Ok(Some(rem.assume_init())),
+            Err(err) => Err(err),
+        }
+    }
+    #[cfg(target_pointer_width = "64")]
     unsafe {
         let mut rem = MaybeUninit::uninit();
-        ret(syscall2(__NR_nanosleep, by_ref(req), out(&mut rem))).map_err(|_err| rem.assume_init())
+        match ret(syscall2(__NR_nanosleep, by_ref(req), out(&mut rem))) {
+            Ok(()) => Ok(None),
+            Err(io::Error::INTR) => Ok(Some(rem.assume_init())),
+            Err(err) => Err(err),
+        }
     }
 }
 
 #[inline]
 pub(crate) fn clock_nanosleep_relative(
     id: __kernel_clockid_t,
-    req: &timespec,
-) -> Result<(), timespec> {
+    req: &__kernel_timespec,
+) -> io::Result<Option<__kernel_timespec>> {
     #[cfg(target_pointer_width = "32")]
     unsafe {
         let mut rem = MaybeUninit::uninit();
-        ret(syscall4(
+        match ret(syscall4(
             __NR_clock_nanosleep_time64,
             clockid_t(id),
             c_int(0),
             by_ref(req),
             out(&mut rem),
         ))
-        .map_err(|_err| rmpt.assume_init())
+        .or_else(|err| {
+            // See the comments in `clock_gettime` about emulation.
+            if err == io::Error::NOSYS {
+                let old_req = __kernel_old_timespec {
+                    tv_sec: req.tv_sec.try_into().map_err(|_| io::Error(EINVAL as _))?,
+                    tv_nsec: req.tv_nsec as _,
+                };
+                let mut old_rem = MaybeUninit::<__kernel_old_timespec>::uninit();
+                let res = ret(syscall4(
+                    __NR_clock_nanosleep,
+                    clockid_t(id),
+                    c_int(0),
+                    by_ref(&old_req),
+                    out(&mut old_rem),
+                ));
+                let old_rem = *old_rem.as_ptr();
+                *rem.as_mut_ptr() = __kernel_timespec {
+                    tv_sec: old_rem.tv_sec.into(),
+                    tv_nsec: old_rem.tv_nsec.into(),
+                };
+                res
+            } else {
+                Err(err)
+            }
+        }) {
+            Ok(()) => Ok(None),
+            Err(io::Error::INTR) => Ok(Some(rem.assume_init())),
+            Err(err) => Err(err),
+        }
     }
     #[cfg(target_pointer_width = "64")]
     unsafe {
         let mut rem = MaybeUninit::uninit();
-        ret(syscall4(
+        match ret(syscall4(
             __NR_clock_nanosleep,
             clockid_t(id),
             c_int(0),
             by_ref(req),
             out(&mut rem),
-        ))
-        .map_err(|_err| rem.assume_init())
+        )) {
+            Ok(()) => Ok(None),
+            Err(io::Error::INTR) => Ok(Some(rem.assume_init())),
+            Err(err) => Err(err),
+        }
     }
 }
 
 #[inline]
-pub(crate) fn clock_nanosleep_absolute(id: __kernel_clockid_t, req: &timespec) -> io::Result<()> {
+pub(crate) fn clock_nanosleep_absolute(
+    id: __kernel_clockid_t,
+    req: &__kernel_timespec,
+) -> io::Result<()> {
     #[cfg(target_pointer_width = "32")]
     unsafe {
         ret(syscall4_readonly(
@@ -1759,6 +1990,24 @@ pub(crate) fn clock_nanosleep_absolute(id: __kernel_clockid_t, req: &timespec) -
             by_ref(req),
             0usize,
         ))
+        .or_else(|err| {
+            // See the comments in `clock_gettime` about emulation.
+            if err == io::Error::NOSYS {
+                let old_req = __kernel_old_timespec {
+                    tv_sec: req.tv_sec.try_into().map_err(|_| io::Error(EINVAL as _))?,
+                    tv_nsec: req.tv_nsec as _,
+                };
+                ret(syscall4_readonly(
+                    __NR_clock_nanosleep,
+                    clockid_t(id),
+                    c_int(0),
+                    by_ref(&old_req),
+                    0usize,
+                ))
+            } else {
+                Err(err)
+            }
+        })
     }
     #[cfg(target_pointer_width = "64")]
     unsafe {
